# $Log: mime.py,v $
# Revision 1.29  2002/05/15 19:26:14  stuart
# More sgmlop / python-1.5 compatibility
#
# Revision 1.28  2002/05/15 18:53:56  stuart
# Work with sgmlop, python1.5 sgmllib.
#
# Revision 1.27  2002/05/15 18:05:26  stuart
# *** empty log message ***
#
# Revision 1.26  2002/05/15 18:03:51  stuart
# Avoid importing declname,declstringlit from sgmllib.  They
# are not present in python-2.2 or sgmlop
#
# Revision 1.25  2002/05/08 19:17:09  stuart
# SGMLFilter fixes and docs.
#
# Revision 1.24  2002/05/06 18:13:31  stuart
# Override parse_declaration in sgmllib.SGMLParser to accept just about
# anything between <! and >
#
# Revision 1.23  2002/05/02 20:31:43  stuart
# Handle quoted-printable HTML attachments.
# Remove entire attachment when HTML can't be parsed by sgmllib.
#
# Revision 1.22  2002/05/02 03:12:40  stuart
# Move check_html to mime module.
#
# Revision 1.21  2002/05/02 02:48:23  stuart
# Remove scripts from HTML even with base64 encoding.
#
# Revision 1.20  2002/05/02 00:21:01  stuart
# Test filtering HTML attachments.
#
# Revision 1.19  2002/05/01 22:12:42  stuart
# Remove scripts from HTML attachments.
#
# Revision 1.18  2002/03/01 20:29:00  stuart
# Ready for release.
#
# Revision 1.17  2002/03/01 19:40:11  stuart
# Compatible mime.defang, and improved mime.check_attachments API
#
# Revision 1.16  2002/03/01 17:35:48  stuart
# Content checking working.
#
# Revision 1.15  2002/02/27 20:07:08  stuart
# Check contents of attachments.
# Report hostname in WARNING.TXT
#
# Revision 1.14  2002/02/22 22:51:31  stuart
# Parse quotes in mime param list
#
# Revision 1.13  2001/12/10 21:05:27  stuart
# Ban more extensions.
#
# Revision 1.12  2001/12/06 00:15:01  stuart
# Handle non-multipart executables.
#
# Revision 1.11  2001/11/27 02:39:45  stuart
# Handle malformed Content-Type header.
#

# This module provides a "defang" function to replace naughty attachments
# with a warning message.

# Author: Stuart D. Gathman <stuart@bmsi.com>
# Copyright 2001 Business Management Systems, Inc.
# This code is under GPL.  See COPYING for details.

# It also provides a MimeMessage class which uses the mimetools and multifile
# modules to parse a MIME message into parts and replace selected parts.  
# There is a more extensive package called mimelib which will be used
# instead in the future.  Currently, mimelib requires the entire message
# to be read into memory for processing.

import StringIO
import rfc822
import multifile
import mimetools
import socket
import Milter
from types import ListType,StringType

def unquote(str):
  """Remove quotes from a string."""
  if len(str) > 1:
      if str[0] == '"':
	if str[-1:] == '"': return str[1:-1]
	try: return str[1:str[1:].index('"')+1]
	except: print str
      if str[0] == '<' and str[-1:] == '>':
	  return str[1:-1]
  return str

def _parseparam(str):
    plist = []
    while str[:1] == ';':
	str = str[1:]
	end = str.find(';')
	while end > 0 and (str.count('"',0,end) & 1):
	  end = str.find(';',end + 1)
	if end < 0: end = len(str)
	f = str[:end]
	if '=' in f:
	    i = f.index('=')
	    f = f[:i].strip().lower() + \
		    '=' + f[i+1:].strip()
	plist.append(f.strip())
	str = str[end:]
    return plist

class MimeHeader:
  "A single MIME header parsed into fields and attributes."
  def __init__(self,str):
    self.typeheader = str
    self.parsetype()
    self.parseplist()

  def parsetype(self):
      str = self.typeheader
      if str is None:
	  str = 'text/plain'
      i = str.find(';')
      if i > 0:
	  self.plisttext = str[i:]
	  str = str[:i]
      else:
	  self.plisttext = ''
      fields = str.split('/')
      for i in range(len(fields)):
	  fields[i] = fields[i].strip().lower()
      self.type = '/'.join(fields)
      self.maintype = fields[0]
      self.subtype = '/'.join(fields[1:])

  def parseplist(self):
      self.plist = _parseparam(self.plisttext)

  def getplist(self):
      return self.plist


  def getparam(self, name):
      name = name.lower() + '='
      n = len(name)
      for p in self.plist:
	  if p[:n] == name:
	      return unquote(p[n:])
      return None

  def getparamnames(self):
      result = []
      for p in self.plist:
	  i = p.find('=')
	  if i >= 0:
	      result.append(p[:i].lower())
      return result

  def getparams(self):
      "Return all parameter names and values."
      result = []
      for p in self.plist:
          i = p.find('=')
	  if i >= 0:
	      result.append((p[:i].lower(),unquote(p[i+1:])))
      return result

  def gettype(self):
      return self.type

class mimepart(multifile.MultiFile):
  def seek(self,pos,whence=0):
    if self.level == 0 and whence == 0:
      self.fp.seek(self.start + pos)
    else:
      multifile.MultiFile.seek(self,pos,whence)

class MimeMessage(mimetools.Message):
  def __init__(self,fp,seekable=1):
    mimetools.Message.__init__(self,fp,seekable)
    self.content_disposition = MimeHeader(self.getheader('content-disposition'))
    self.content_type = MimeHeader(self.getheader('content-type'))
    self._payload = None
    self.headerchange = None
    self.modified = 0

  # working getparam
  def getparam(self,name):
    return self.content_type.getparam(name)

  def __setitem__(self, name, value):
    rc = mimetools.Message.__setitem__(self,name,value)
    self.modified = 1
    if self.headerchange: self.headerchange(self,name,value)
    return rc

  def __delitem__(self, name):
    if self.headerchange: self.headerchange(self,name,None)
    rc = mimetools.Message.__delitem__(self,name)
    self.modified = 1
    return rc

  def getname(self):
    return self.getparam('name')

  def getnames(self):
    """Return a list of (attr,name) pairs of attributes that IE might
       interpret as a name - and hence decide to execute this message."""
    return self.content_type.getparams() + [("filename",self.getfilename())]

  def getfilename(self):
    return self.content_disposition.getparam('filename')

  def _parsebody(self):
    self.rewindbody()
    #print self.headers
    #print self.fp.read()
    #self.rewindbody()
    msgtype = self.gettype()
    file = multifile.MultiFile(self.fp)
    self.boundary = boundary = self.content_type.getparam("boundary")
    if not boundary or not msgtype.startswith('multipart/'):
      return self.fp
    list = []
    file.push(boundary)
    self.preamble = file.read()
    try:
      while file.next():
	subfile = mimepart(self.fp)
	subfile.push(boundary)
	submsg = MimeMessage(subfile)
	list.append(submsg)
    except:
      if len(list) > 0: del list[-1]	# missing end boundary
    file.pop()
    self.epilog = file.read()
    return list

  def ismultipart(self):
    if not self._payload:
      msgtype = self.gettype()
      boundary = self.getparam("boundary")
      return boundary and msgtype.startswith('multipart/')
    return type(self.get_payload()) is ListType

  def ismodified(self):
    "True if this message or a subpart has been modified."
    if not self.ismultipart(): return self.modified
    for i in self._payload:
      if i.ismodified(): return 1
    return 0

  #
  # Payload manipulation.
  #
  def add_payload(self, payload):
      """Add the given payload to the current payload.

      If the current payload is empty, then the current payload will be made
      a scalar, set to the given value.
      """
      if self._payload is None:
	  self._payload = payload
      elif type(self._payload) is ListType:
	  self._payload.append(payload)
      elif self.getmaintype() not in (None, 'multipart'):
	  raise Exception(
	      'Message main Content-Type: must be "multipart" or missing')
      else:
	  self._payload = [self._payload, payload]
      self.modified = 1

  def get_payload(self, i=None):
      """Return the current payload exactly as is.

      Optional i returns that index into the payload.
      """
      if not self._payload:
	self._payload = self._parsebody()
      if i is None:
	  if self._payload is self.fp:
	    self.rewindbody()
	  elif type(self._payload) is not ListType:
	    self._payload.seek(0)
	  return self._payload
      if type(self._payload) is not ListType:
	  raise TypeError, i
      return self._payload[i]

  def set_payload(self, payload):
      """Set the payload to the given value."""
      if type(payload) is StringType:
        self._payload = StringIO.StringIO(payload)
      else:
	self._payload = payload
      self.modified = 1

  def dump(self,file):
    file.write(''.join(self.headers))
    file.write('\n')
    body = self._payload
    if type(body) == type([]):
      boundary = '--' + self.boundary + '\n'
      if self.preamble:
	file.write(self.preamble)
      for part in body:
	file.write(boundary)
	part.dump(file)
      file.write('--' + self.boundary + '--\n')
      if self.epilog:
	file.write(self.epilog)
    else:
      file.write(self.get_payload().read())

extlist = ''.join("""
ade,adp,asd,asx,bas,bat,chm,cmd,com,cpl,crt,dll,exe,hlp,hta,inf,ins,isp,js,jse,
lnk,mdb,mde,msc,msi,msp,mst,ocx,pcd,pif,reg,scr,sct,shs,url,vb,vbe,vbs,wsc,
wsf,wsh 
""".split())
bad_extensions = map(lambda x:'.' + x,extlist.split(','))

def check_ext(name):
  if not name: return name
  lname = name.lower()
  for ext in bad_extensions:
    if lname.endswith(ext): return name
  return None

virus_msg = """This message appeared to contain a virus.
It was originally named '%s', and has been removed.
A copy of your original message was saved as '%s:%s'.
See your administrator.
"""

def check_name(msg,savname=None,ckname=check_ext):
  "Replace attachment with a warning if its name is suspicious."
  for (key,name) in msg.getnames():
    badname = ckname(name)
    if badname:
      hostname = socket.gethostname()
      msg.set_payload(virus_msg % (badname,hostname,savname))
      del msg["content-type"]
      del msg["content-disposition"]
      del msg["content-transfer-encoding"]
      name = "WARNING.TXT"
      msg["Content-Type"] = "text/plain; name="+name
      break
  return Milter.CONTINUE

def check_attachments(msg,check):
  """Scan attachments.
msg	MimeMessage
check	function(MimeMessage): int
	Return CONTINUE, REJECT, ACCEPT
  """
  if msg.ismultipart():
    for i in msg.get_payload():
      rc = check_attachments(i,check)
      if rc != Milter.CONTINUE: return rc
    return Milter.CONTINUE
  return check(msg)

# save call context for Python without nested_scopes
class _defang:
  def __init__(self,savname,check):
    self._savname = savname
    self._check = check
  def _chk_name(self,msg):
    return check_name(msg,self._savname,self._check)

# emulate old defang function
def defang(msg,savname=None,check=check_ext):
  "Compatible entry point."
  check_attachments(msg,_defang(savname,check)._chk_name)
  if msg.ismodified():
    return 1;
  return 0

import sgmllib
import base64
import quopri

import re
declname = re.compile(r'[a-zA-Z][-_.a-zA-Z0-9]*\s*')
declstringlit = re.compile(r'(\'[^\']*\'|"[^"]*")\s*')

class SGMLFilter(sgmllib.SGMLParser):
  """Parse HTML and pass through all constructs unchanged.  It is intended for
   derived classes to implement exceptional processing for selected cases."""

  def __init__(self,out):
    sgmllib.SGMLParser.__init__(self)
    self.out = out

  def handle_comment(self,comment):
    self.out.write("<!--%s-->" % comment)

  def unknown_starttag(self,tag,attr):
    if hasattr(self,"get_starttag_text"):
      self.out.write(self.get_starttag_text())
    else:
      self.out.write("<%s" % tag)
      for (key,val) in attr:
        self.out.write(' %s="%s"' % (key,val))
      self.out.write('>')

  def handle_data(self,data):
    self.out.write(data)

  def handle_entityref(self,ref):
    self.out.write("&%s;" % ref)

  def handle_charref(self,ref):
    self.out.write("&#%s;" % ref)
      
  def unknown_endtag(self,tag):
    self.out.write("</%s>" % tag)

  def handle_special(self,data):
    self.out.write("<!%s>" % data)

  def write(self,buf):
    "Act like a writer.  Why doesn't SGMLParser do this by default?"
    self.feed(buf)

  # Python-2.1 sgmllib rejects illegal declarations.  Since various Microsoft
  # products accept and output them, we need to pass them through -
  # at least until we discover that MS will execute them.
  # sgmlop-1.1 will not use this method, but calls handle_special to
  # do what we want.
  def parse_declaration(self, i):
      rawdata = self.rawdata
      j = i + 2
      while 1:
	  c = rawdata[j]
	  if c == ">":
	      # end of declaration syntax
	      self.handle_special(rawdata[i+2:j])
	      return j + 1
	  if c in "\"'":
	      m = declstringlit.match(rawdata, j)
	      if not m:
		  # incomplete or an error?
		  return -1
	      j = m.end()
	  elif c in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ":
	      m = declname.match(rawdata, j)
	      if not m:
		  # incomplete or an error?
		  return -1
	      j = m.end()
	  elif j == len(rawdata):
	      # end of buffer between tokens
	      return -1
	  else:
	      j += 1
      assert 0, "can't get here!"

class HTMLScriptFilter(SGMLFilter):
  "Remove scripts from an HTML document."
  def __init__(self,out):
    SGMLFilter.__init__(self,out)
    self.ignoring = 0
    self.modified = 0
    self.msg = "<!-- WARNING: embedded script removed -->"
  def start_script(self,unused):
    self.ignoring += 1
    self.modified = 1
    self.out.write(self.msg)
  def end_script(self):
    self.ignoring -= 1
  def handle_data(self,data):
    if not self.ignoring: SGMLFilter.handle_data(self,data)
  def handle_comment(self,comment):
    if not self.ignoring: SGMLFilter.handle_comment(self,comment)

def check_html(msg,savname=None):
  "Remove scripts from HTML attachments."
  msgtype = msg.gettype().lower()
  # check for more MSIE braindamage
  if msgtype == 'application/octet-stream':
    for (attr,name) in msg.getnames():
      if name and name.lower().endswith(".htm"):
	msgtype = 'text/html'
  if msgtype == 'text/html':
    fp = msg.get_payload()
    out = StringIO.StringIO()
    filter = HTMLScriptFilter(out)
    enc = msg.getencoding()
    try:
      if enc == 'base64':
	base64.decode(fp,filter)
      elif enc == 'quoted-printable':
	quopri.decode(fp,filter)
      else:
	enc = None
	mimetools.copyliteral(fp,filter)
      filter.close()
    #except sgmllib.SGMLParseError:
    except RuntimeError:
      #mimetools.copyliteral(msg.get_payload(),open('debug.out','w')
      hostname = socket.gethostname()
      msg.set_payload(
  "An HTML attachment could not be parsed.  The original is saved in %s:%s"
      % (hostname,savname))
      del msg["content-type"]
      del msg["content-disposition"]
      del msg["content-transfer-encoding"]
      name = "WARNING.TXT"
      msg["Content-Type"] = "text/plain; name="+name
      return Milter.CONTINUE
    if filter.modified:
      if enc:
	fp = out
	out = StringIO.StringIO()
	fp.seek(0)
	quopri.encode(fp,out,1)
      msg.set_payload(out)	# remove embedded scripts
  return Milter.CONTINUE
